package EM_LLR_Utils;



import java.util.Random;

import edu.tau.compbio.ds.SimilarityMatrix;
import edu.tau.compbio.stat.NormalDistribution;

public abstract class AbstractParamsEM {
	protected NormalDistribution _matesDist = null;
	protected NormalDistribution _nonDist = null;
	protected double _matesP;
	protected double _likelihood;	
	
	//protected float[][] _sims = null;
	protected double[] _regPs = null;
	protected SimilarityMatrix<String> _simMat = null;

	protected int _inpSize = 0;

	protected boolean _updateExp = true;
	protected boolean _updateStd = true;
	protected boolean _updateP = true;

	protected int[] _inds = null;

	protected int _nonNaNCount = 0;
	protected double _regsSum = 0;
	
	protected boolean _selfInit = true;
	protected boolean _sameStd = false;
	
	protected final static int MIN_VALS_NUM = 5000;
	protected final double EPS_IMPROVE = 100;
	protected final double INIT_MATES_PROB = 0.01;

	protected int _plotIterCount = 1;

	public AbstractParamsEM(SimilarityMatrix<String> simMat,double[] regPs)
	{
		_inds = new int[simMat.size()];
		for (int k=0;k<_inds.length;k++)
			_inds[k] = k;
		init(simMat,regPs);
	}
	public AbstractParamsEM(SimilarityMatrix<String> simMat,double[] regPs,int randCount)
	{
		_inpSize = randCount;
		boolean[] used = new boolean[simMat.size()];
		if (randCount > simMat.size())
			throw new IllegalStateException("The random count exceeds the number of elements");
		_inds = new int[_inpSize];
		Random rand = new Random();
		for (int k=0;k<_inds.length;k++)
		{
			int ind = rand.nextInt(simMat.size());
			while (used[ind])
			{
				ind = rand.nextInt(simMat.size());
			}
			_inds[k] = ind;
			used[ind] = true;
		}
		init(simMat,regPs);
	}
	public void setUpdateMode(boolean updateExp,boolean updateStd,boolean updateP)
	{
		_updateExp = updateExp;
		_updateStd = updateStd;
		_updateP = updateP;
	}
	public void init(SimilarityMatrix<String> simMat,double[] regPs)
	{
		_simMat = simMat;
		_inpSize = _inds.length;
		//_sims = simMat.getSimilarities();
		_regPs = regPs;
		initNonNaNCount();
	}


	public NormalDistribution getMatesDistribution()
	{
		return _matesDist;
	}
	
	public NormalDistribution getNonMatesDistribution()
	{
		return _nonDist;
	}
	
	public double getMatesP()
	{
		return _matesP;
	}
	
	public double getLikelihood()
	{
		return _likelihood;
	}
	public void setSameStd(boolean flag)
	{
		_sameStd = flag;
	}

	protected void initNonNaNCount()
	{
		_nonNaNCount = 0;
		_regsSum = 0;
		for (int i=0;i<_simMat.size();i++)
		{
			for (int j=i+1;j<_simMat.size();j++)
			{
				if (!Float.isNaN(_simMat.getSimilarity(_inds[i], _inds[j])))
				{
					_nonNaNCount++;
					_regsSum += _regPs[_inds[i]]*_regPs[_inds[j]];
				}
			}
		}
		
		System.out.println("EM will work with "+_nonNaNCount+" non-NaN elements");
	}

//////////////////////////////////////////////////////////////////////////////////////////////
//	Func Name  : CalcExp
//	Parameters : g_i - case i g values vector
//				   sumG_i - case i g values sum
//	            
//	Description: computes the maximum expectation for case i
//	Returns    : the computed expectation
////////////////////////////////////////////////////////////////////////////////////////////	//
	protected double calcExp(double[] g_i, double sumG_i)
	{
		double sigmaGS = 0;
		int count = 0;
		for (int i=0;i<_inpSize;i++)
		{
			for (int j=i+1;j<_inpSize;j++)
			{
				float sim = this._simMat.getSimilarity(_inds[i], _inds[j]);//_sims[_inds[i]][_inds[j]];
				if (Float.isNaN(sim))
					continue;
				sigmaGS += g_i[count++] * sim;
			}
		}
		if (count!=g_i.length)
		{
			throw new IllegalStateException("The vector size is bogus!");
		}
		return sigmaGS/sumG_i;
	}

//////////////////////////////////////////////////////////////////////////////////////////////
//	Func Name  : CalcStd
//	Parameters : g_i - case i g values vector
//				   sumG_i - case i g values sum
//	             exp - case i maximum expectation
//	             simVals - the similarity values vector
//	Description: computes the maximum standard deviation for case i
//	Returns    : the computed std
////////////////////////////////////////////////////////////////////////////////////////////	//
	protected double calcStd(double[] g_i, double sumG_i, double exp)
	{
		double sigma = 0;
		int count = 0;
		for (int i=0;i<_inpSize;i++)
		{
			for (int j=i+1;j<_inpSize;j++)
			{
				float sim = this._simMat.getSimilarity(_inds[i], _inds[j]);//__sims[_inds[i]][_inds[j]];
				if (Float.isNaN(sim))
					continue;
				sigma += (g_i[count++] * (exp - sim)*(exp - sim));
			}
		}
		if (count != g_i.length)
		{
			throw new IllegalStateException("The vector size is bogus!");
		}
		return Math.sqrt(sigma/sumG_i);
	}

}
