package EM_LLR_Utils;

import java.util.ArrayList;
import java.util.Arrays;

import edu.tau.compbio.ds.SimilarityMatrix;
import edu.tau.compbio.stat.NormalDistribution;
import edu.tau.compbio.util.ProgressManager;

public class LLRScoresCalculator {

	ArrayList<String> ids;
	float[][] similarities;

	double[][] weightedMatrix;

	float matesMaxP = 0.15f;
	boolean init = false;
	private NormalDistribution matesDist;
	private NormalDistribution nonMatesDist;
	private double sigmaF;
	private double sigmaT;
	private double sigmaSqT;
	private double sigmaSqF;
	private double meanF;
	private double meanT;
	private double oneMinusP;
	private double pMates;

	public LLRScoresCalculator(ArrayList<String> names, float[][] sims) {
		ids = names;
		similarities = sims;
	}

	public LLRScoresCalculator(ArrayList<String> names, double[][] sims) {
		ids = names;
		similarities = new float[sims.length][sims.length];
		for (int i = 0; i < sims.length; i++) {
			for (int j = 0; j < sims.length; j++) {
				similarities[i][j] = (float) sims[i][j];
			}
		}
	}

	public LLRScoresCalculator(float[][] sims) {
		similarities = sims;
	}

	public double[][] getClickWeightedMatrix(boolean doEM) {

		if (doEM)
			runEM();

		for (int i = 0; i < weightedMatrix.length; i++) {
			for (int j = 0; j < i; j++) {
				double sim = similarities[i][j];

				double score = getSimLikeLihood(sim,false);

				weightedMatrix[i][j] = score;
				weightedMatrix[j][i] = weightedMatrix[i][j];

				// if (j==0){
				// System.out.println(sim+","+likelihood+","+score+"\t"+dev1+"\t"+dev2);
				// }

			}
		}

		return weightedMatrix;
	}

	public double getSimLikeLihood(double sim,boolean print) {
		double score = Math.log((pMates * sigmaF) / (oneMinusP * sigmaT));
		if (print){
			System.out.println("factor constant: " + score);
			System.out.println("deviation from mates: " + (Math.pow((sim - meanT), 2) / sigmaSqT));
			System.out.println("deviation from non mates: " + (Math.pow((sim - meanF), 2) / sigmaSqF));
		}
		score += (Math.pow((sim - meanF), 2) / sigmaSqF);
		score -= (Math.pow((sim - meanT), 2) / sigmaSqT);
		return score;
	}

	public double getMeanMatesLikelihood() {
		return getSimLikeLihood(meanT,false);
	}

	public void runEM() {
		System.out.println("Executing EM...");
		double[] regs = new double[similarities.length];
		Arrays.fill(regs, 1f);

		boolean _updateExp = true;
		boolean _updateStd = true;
		boolean _updateP = true;

		float _emMinPMates = 0.01f;
		float _emMaxPMates = matesMaxP;

		// SimpleParamsEM em = new SimpleParamsEM(new
		// SimilarityMatrix<String>(ids, similarities),regs);
		RegParamsEM em = new RegParamsEM(new SimilarityMatrix<String>(ids,
				similarities), regs);
		em.setUpdateMode(_updateExp, _updateStd, _updateP);
		em.setPMatesRange(_emMinPMates, _emMaxPMates);
		em.setSameStd(false);
		if (init) {
			em.setInit(true);
		}
		ProgressManager progMan = new ProgressManager();
		em.run(progMan);
		matesDist = em.getMatesDistribution();
		nonMatesDist = em.getNonMatesDistribution();
		pMates = em.getMatesP();
		weightedMatrix = new double[ids.size()][ids.size()];

		sigmaF = nonMatesDist.getStd();
		sigmaT = matesDist.getStd();

		sigmaSqT = sigmaT * sigmaT * 2;
		sigmaSqF = sigmaF * sigmaF * 2;
		meanF = nonMatesDist.getExp();
		meanT = matesDist.getExp();

		oneMinusP = 1 - pMates;
	}

	// public LinkedList<LinkedList<String>> getClusters() {
	// getClickWeightedMatrix();
	// LinkedList<LinkedList<Integer>> singletons = new
	// LinkedList<LinkedList<Integer>>();
	// for (int i = 0; i < weightedMatrix.length; i++) {
	// LinkedList<Integer> l = new LinkedList<Integer>();
	// l.add(i);
	// singletons.add(l);
	// }
	// LinkedList<LinkedList<Integer>> clustersIndices = greedyImprovements(
	// singletons, 0, 10);
	// LinkedList<LinkedList<String>> clusters =
	// getClustersBuNames(clustersIndices);
	//
	// return clusters;
	// }
	//
	// private LinkedList<LinkedList<String>> getClustersBuNames(
	// LinkedList<LinkedList<Integer>> clustersIndices) {
	// LinkedList<LinkedList<String>> clusters = new
	// LinkedList<LinkedList<String>>();
	//
	// for (Iterator<LinkedList<Integer>> iterator = clustersIndices
	// .iterator(); iterator.hasNext();) {
	// LinkedList<Integer> clusterByIndices = (LinkedList<Integer>) iterator
	// .next();
	//
	// LinkedList<String> clust = new LinkedList<String>();
	// for (Iterator<Integer> iterator2 = clusterByIndices.iterator(); iterator2
	// .hasNext();) {
	// Integer index = (Integer) iterator2.next();
	// clust.add(ids.get(index));
	// }
	//
	// clusters.add(clust);
	//
	// }
	// return clusters;
	// }

	// public LinkedList<LinkedList<Integer>> getSeeds() {
	//
	// int[] indices = new int[weightedMatrix.length];
	// for (int i = 0; i < indices.length; i++) {
	// indices[i]=i;
	// }
	//
	// LinkedList<LinkedList<Integer>> seeds =
	// StaticGraphFunctions.getStupidMaxCut(indices, matrixForSeeds,-1);
	// System.out.println("calculated cut");
	// LinkedList<LinkedList<Integer>> finalSeeds = new
	// LinkedList<LinkedList<Integer>>();
	//
	// while (seeds.size() > 0){
	// LinkedList<Integer> currentSeed = seeds.removeFirst();
	//
	// if (currentSeed.size() < 5){
	// finalSeeds.add(currentSeed);
	// continue;
	// }
	//
	// int[] currentSeedIndices = intListToArray(currentSeed);
	//
	// // check if seed is valid
	// double sum = 0;
	// for (int i = 0; i < currentSeedIndices.length; i++) {
	// for (int j = 0; j < i; j++) {
	// sum+=weightedMatrix[currentSeedIndices[i]][currentSeedIndices[j]];
	// }
	// }
	//
	// if (sum > 0){
	// System.out.println("accepted a seed of size: " + currentSeed.size());
	// finalSeeds.add(currentSeed);
	// }
	// else{
	// LinkedList<LinkedList<Integer>> cut =
	// StaticGraphFunctions.getStupidMaxCut(currentSeedIndices,
	// matrixForSeeds,-1);
	// // System.out.println("calculated cut");
	// seeds.add(cut.get(0));
	// seeds.add(cut.get(1));
	// }
	//
	// }
	//
	// return finalSeeds;
	// }
	//
	// private int[] intListToArray(LinkedList<Integer> currentSeed) {
	// int[] currentSeedIndices = new int[currentSeed.size()];
	// int i=0;
	// for (Iterator<Integer> iterator = currentSeed.iterator();
	// iterator.hasNext();) {
	// Integer index = (Integer) iterator.next();
	// currentSeedIndices[i]=index;
	// i++;
	// }
	// return currentSeedIndices;
	// }
	//
	// /**
	// * This is a simple agglomerative step. Iteratively, choose the best pair
	// * and unite if the cut probability is above the given threshold.
	// Threshold
	// * default is 0
	// *
	// * @param seeds
	// */
	// public LinkedList<LinkedList<Integer>> greedyImprovements(
	// LinkedList<LinkedList<Integer>> seeds, double thr, int minSize) {
	//
	// System.out.println("improving clustering ");
	// if (seeds.size() < 3) {
	// throw new IllegalArgumentException();
	// }
	//
	// // initialize the weighted seeds similarity matrix
	// double[][] sumMatrix = new double[seeds.size()][seeds.size()];
	//
	// int seedsFirstCounter = 0;
	// for (Iterator<LinkedList<Integer>> iterator = seeds.iterator(); iterator
	// .hasNext();) {
	// LinkedList<Integer> firstSeed = (LinkedList<Integer>) iterator
	// .next();
	//
	// int seedsSecondCounter = 0;
	// for (Iterator<LinkedList<Integer>> iterator2 = seeds.iterator();
	// iterator2
	// .hasNext();) {
	// LinkedList<Integer> secondSeed = (LinkedList<Integer>) iterator2
	// .next();
	// if (firstSeed == secondSeed) {
	// break;
	// }
	//
	// // go over the cut between the seeds
	// for (Iterator<Integer> iterator3 = secondSeed.iterator(); iterator3
	// .hasNext();) {
	// Integer firstGeneIndex = (Integer) iterator3.next();
	//
	// for (Iterator<Integer> iterator4 = firstSeed.iterator(); iterator4
	// .hasNext();) {
	// Integer secondGeneIndex = (Integer) iterator4.next();
	//
	// sumMatrix[seedsFirstCounter][seedsSecondCounter] +=
	// weightedMatrix[firstGeneIndex][secondGeneIndex];
	// }
	// }
	//
	// seedsSecondCounter++;
	//
	// }
	//
	// seedsFirstCounter++;
	// }
	//
	// // set the current max cut
	// double maxCut = 0;
	// int maxFirstIndex = -1;
	// int maxSecondIndex = -1;
	// for (int i = 0; i < sumMatrix.length; i++) {
	// for (int j = 0; j < i; j++) {
	// sumMatrix[j][i] = sumMatrix[i][j];
	// if (maxCut < sumMatrix[i][j]) {
	// maxCut = sumMatrix[i][j];
	// maxFirstIndex = i;
	// maxSecondIndex = j;
	// }
	// }
	// }
	//
	// // here we merge seeds
	// // always merge into the smallest index
	// // in this step, change the weights matrix
	//
	// // this is an auxiliary boolean array
	// boolean[] isValidCluster = new boolean[seeds.size()];
	// Arrays.fill(isValidCluster, true);
	//
	// while (maxCut > thr) {
	//
	// System.out.println("merging seeds max cut size: " + maxCut);
	// // 1. merge the relevant lists
	// LinkedList<Integer> first = seeds.get(maxFirstIndex);
	// LinkedList<Integer> second = seeds.get(maxSecondIndex);
	//
	// for (Iterator<Integer> iterator = second.iterator(); iterator
	// .hasNext();) {
	// Integer geneIndex = (Integer) iterator.next();
	// first.add(geneIndex);
	// }
	//
	// second = new LinkedList<Integer>();
	// isValidCluster[maxSecondIndex] = false;
	//
	// // update the weights matrix
	// for (int i = 0; i < sumMatrix.length; i++) {
	// if (!isValidCluster[i]) {
	// continue;
	// }
	// sumMatrix[i][maxFirstIndex] += sumMatrix[i][maxSecondIndex];
	// sumMatrix[maxFirstIndex][i] = sumMatrix[i][maxFirstIndex];
	// }
	//
	// // update the maximal cut
	// maxCut = 0;
	// maxFirstIndex = -1;
	// maxSecondIndex = -1;
	//
	// for (int i = 0; i < sumMatrix.length; i++) {
	// if (!isValidCluster[i]) {
	// continue;
	// }
	// for (int j = 0; j < i; j++) {
	// if (!isValidCluster[j]) {
	// continue;
	// }
	// sumMatrix[j][i] = sumMatrix[i][j];
	// if (maxCut < sumMatrix[i][j]) {
	// maxCut = sumMatrix[i][j];
	// maxFirstIndex = i;
	// maxSecondIndex = j;
	// }
	// }
	// }
	// }
	//
	// LinkedList<LinkedList<Integer>> clusters = new
	// LinkedList<LinkedList<Integer>>();
	// for (Iterator<LinkedList<Integer>> iterator = seeds.iterator(); iterator
	// .hasNext();) {
	// LinkedList<Integer> seed = (LinkedList<Integer>) iterator.next();
	// System.out.println("seed size: " + seed.size());
	// if (seed.size() >= minSize) {
	// clusters.add(seed);
	// }
	// }
	//
	// return clusters;
	//
	// }
	//
	// public LinkedList<LinkedList<Integer>> doHierarchicalClustering(){
	// LinkedList<LinkedList<Integer>> seeds = new
	// LinkedList<LinkedList<Integer>>();
	// for (int i = 0; i < weightedMatrix.length; i++) {
	// LinkedList<Integer> singleton = new LinkedList<Integer>();
	// singleton.add(i);
	// seeds.add(singleton);
	//
	// }
	// return greedyImprovements(seeds, 0, 1);
	// }

	public double[][] getWeightedMatrix() {
		return weightedMatrix;
	}

	public void setMatesMaxP(float matesMaxP) {
		this.matesMaxP = matesMaxP;
		init = true;
	}

	public void setDistributions(NormalDistribution mates,
			NormalDistribution nonMates, double pMates) {
		matesDist = mates;
		nonMatesDist = nonMates;
		this.pMates = pMates;

		weightedMatrix = new double[ids.size()][ids.size()];

		sigmaF = nonMatesDist.getStd();
		sigmaT = matesDist.getStd();

		sigmaSqT = sigmaT * sigmaT * 2;
		sigmaSqF = sigmaF * sigmaF * 2;
		meanF = nonMatesDist.getExp();
		meanT = matesDist.getExp();

		oneMinusP = 1 - pMates;
	}
	
}
