package graph_algorithms;

import java.util.HashSet;
import java.util.Iterator;

/**
 * 
 * This class implements a simple algorithm that tries to merge two meta-modules.
 * If the two meta-modules can be merged then the result is the new meta-module.
 * Otherwise the result is null.
 *
 */

public class MetaModulePairMerger {
	
	/* Members */
	MetaModule newPair;
	double averageBetweenAddedDCScore;
	double averageWithinAddedGCScore;
	static final double  sizeRationsThr = 0.25;
	
	public MetaModulePairMerger() {
	}

	public static MetaModulePairMerger merge(MetaModule a, MetaModule b){
		MetaModulePairMerger m1 = new MetaModulePairMerger();
		MetaModulePairMerger m2 = new MetaModulePairMerger();
		MetaModule try1 = m1.mergeLists(a.U,b.U,a.V,b.V,a.dcData);
		MetaModule try2 = m2.mergeLists(a.U, b.V, a.V, b.U,a.dcData);
		
		if (try1==null && try2==null){
			return null;
		}
		if(try1!=null && try2!=null){
			if (m1.averageBetweenAddedDCScore>m2.averageBetweenAddedDCScore){
				return m1;
			}
			else{
				return m2;
			}
		}
		if (try1!=null){
			return m1;
		}
		if (try2!=null){
			return m2;
		}
		return null;
	}
	
	public MetaModule mergeLists(HashSet<Integer> u1,HashSet<Integer> u2,
			HashSet<Integer> v1, HashSet<Integer> v2,
			MetaModuleFinder data) {

//		System.out.print("Trying to merge, ");
		HashSet<Integer> u1PlusU2 = new HashSet<Integer>(u1);
		u1PlusU2.addAll(u2);
		HashSet<Integer> v1PlusV2 = new HashSet<Integer>(v1);
		v1PlusV2.addAll(v1);
		double n1 = u1PlusU2.size();
		double n2 = v1PlusV2.size();
		double ratio = Math.min(n1/n2,n2/n1);
		if (ratio<sizeRationsThr){
//			System.out.println("failed, new ratio is " + ratio);
			return null;
		}
		
		
		double u2uScore = getBetweenGCScores(u1, u2,data);
		if (u2uScore<=0){
//			System.out.println("failed, u2u score is " + u2uScore);
			return null;
		}

		
		double v2vScore = getBetweenGCScores(v1, v2, data);
		if (v2vScore<=0){
//			System.out.println("failed, v2v score is  " + v2vScore);
			return null;
		}
		
		double u1Tov2Score = getBetweenDCScores(u1, v2, data);
		if (u1Tov2Score<=0){
//			System.out.println("failed, u1 to v2 score is : " + u1Tov2Score);
			return null;
		}
		double u2Tov1Score = getBetweenDCScores(u2, v1, data);
		if (u2Tov1Score<=0){
//			System.out.println("failed, u2 to v1 score is : " + u1Tov2Score);
			return null;
		}
		
		averageBetweenAddedDCScore=0;
		averageBetweenAddedDCScore += u1Tov2Score/((double)u1.size()*v2.size());
		averageBetweenAddedDCScore += u2Tov1Score/((double)u2.size()*v1.size());
		System.out.println("SUCCESS!!!!: new average score is " + averageBetweenAddedDCScore);
		
		averageWithinAddedGCScore=0;
		averageWithinAddedGCScore += u2uScore/((double)u1.size()*u2.size());
		averageWithinAddedGCScore += v2vScore/((double) v1.size()*v2.size());
		
		newPair = new MetaModule(u1PlusU2, v1PlusV2, data); 
		return newPair;
	}
	
	/**
	 * 
	 * A set of auxiliary static methods
	 * 
	 */

	public static double getBetweenGCScores(HashSet<Integer> a,
			HashSet<Integer> b, MetaModuleFinder data) {

		double score = 0;
		for (Iterator<Integer> iterator = b.iterator(); iterator.hasNext();) {
			Integer inB = (Integer) iterator.next();
			for (Iterator<Integer> iterator2 = a.iterator(); iterator2
					.hasNext();) {
				Integer inA = (Integer) iterator2.next();
				if (inA!=inB){
					score += data.consistentCorrelationsGraph[inA][inB];
				}
			}

		}
		return score;
	}

	public static double getBetweenAvgGCScores(HashSet<Integer> a,
			HashSet<Integer> b, MetaModuleFinder data) {

		double score = 0;
		for (Iterator<Integer> iterator = b.iterator(); iterator.hasNext();) {
			Integer inB = (Integer) iterator.next();
			for (Iterator<Integer> iterator2 = a.iterator(); iterator2
					.hasNext();) {
				Integer inA = (Integer) iterator2.next();
				if (inA!=inB){
					score += data.consistentCorrelationsGraph[inA][inB];
				}
			}

		}
		return score / ((double) b.size() + a.size());
	}

	public static double getBetweenDCScores(HashSet<Integer> a,
			HashSet<Integer> b, MetaModuleFinder data) {

		double score = 0;
		for (Iterator<Integer> iterator = b.iterator(); iterator.hasNext();) {
			Integer inB = (Integer) iterator.next();
			for (Iterator<Integer> iterator2 = a.iterator(); iterator2
					.hasNext();) {
				Integer inA = (Integer) iterator2.next();
				score += data.DC_graph[inA][inB];
			}

		}
		return score;
	}
	
	public static double getBetweenScores(HashSet<Integer> a,
			HashSet<Integer> b, double[][] mat) {

		double score = 0;
		for (Iterator<Integer> iterator = b.iterator(); iterator.hasNext();) {
			Integer inB = (Integer) iterator.next();
			for (Iterator<Integer> iterator2 = a.iterator(); iterator2
					.hasNext();) {
				Integer inA = (Integer) iterator2.next();
				if (inA==inB){
					continue;
				}
				score += mat[inA][inB];
			}

		}
		return score;
	}

	public static double getBetweenAvgDCScores(HashSet<Integer> a,
			HashSet<Integer> b, MetaModuleFinder data) {

		double score = 0;
		for (Iterator<Integer> iterator = b.iterator(); iterator.hasNext();) {
			Integer inB = (Integer) iterator.next();
			for (Iterator<Integer> iterator2 = a.iterator(); iterator2
					.hasNext();) {
				Integer inA = (Integer) iterator2.next();
				score += data.DC_graph[inA][inB];
			}

		}
		return score / ((double) b.size() + a.size());
	}
}
