package main_scripts;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;

import math_utils.ExpressionMatrix;
import math_utils.staticMathHelperMethods;
import edu.tau.compbio.math.TailProbs;
import edu.tau.compbio.stat.NormalDistribution;
import EM_LLR_Utils.LLRScoresCalculator;

public class DCGraphsCalculator {
	
	protected double K=2;
	
	// Memebers for the statistical class-specific analysis
	protected TscoresCalculator tScoresCalc;
	
	protected float[][] realTscoresGraph;
	protected float[] randomScores;
	protected LLRScoresCalculator llrAnalysisForTscores;
	protected LLRScoresCalculator globalCorrelationsLLRAnalyzer;

	// The results of the statistical analysis: up DC, down DC and the CG graphs
	protected double[][] dGGraph;
	protected double[][] uGGraph;
	protected double[][] consistentCorrelationsGraph;
	protected int classIndex;
	
	public DCGraphsCalculator(TscoresCalculator t, int classI) {
		tScoresCalc = t;
		classIndex =classI;
	}
	
	
	public void doStatisticalAnalysis() {
		getDCGraphs();
		getConsistentCorrelationsGraph();
	}


	public void getDCGraphs() {
		tScoresCalc.setGlobalGraphForOneClass(classIndex);
		realTscoresGraph = tScoresCalc.scoredTscoresGraph;
		float[] realScores = new float[realTscoresGraph.length
				* (realTscoresGraph.length - 1) / 2];
		int count = 0;
		for (int i = 0; i < realTscoresGraph.length; i++) {
			for (int j = 0; j < i; j++) {
				realScores[count] = realTscoresGraph[i][j];
				count++;
			}
		}

		float realAvg = staticMathHelperMethods.Average(realScores);
		float realSD = staticMathHelperMethods.sStatistic(realScores, realAvg);

		Runtime.getRuntime().gc();

		// get clustering
		// first get random dca scores
		int randSize = 500;
		int numGenes = tScoresCalc.data.matrix.length;
		LinkedList<Float> randScores = new LinkedList<Float>();
		Random rand = new Random();
		for (int i = 0; i < 50; i++) {

			// get subMatrix
			LinkedList<Integer> rowIndices = new LinkedList<Integer>();
			for (int j = 0; j < randSize; j++) {
				int nextInt = rand.nextInt(numGenes);
				while (rowIndices.contains(nextInt)) {
					nextInt = rand.nextInt(numGenes);
				}
				rowIndices.add(nextInt);
			}

			ExpressionMatrix subMatrix = tScoresCalc.data
					.getSubMatrixByRowsIndices(rowIndices);
			ExpressionMatrix permutedMatrix = subMatrix
					.getShuffeledMatrixByColumns(rand);

			TscoresCalculator shuffComp = new TscoresCalculator(permutedMatrix,
					tScoresCalc.classesBorders, tScoresCalc.ClassesNames,
					"pearson");
			shuffComp.setGlobalGraphForOneClass(classIndex);
			float[][] matrix = shuffComp.scoredTscoresGraph;
			for (int j = 0; j < matrix.length; j++) {
				for (int j2 = 0; j2 < j; j2++) {
					randScores.add(matrix[j][j2]);
				}
			}
			
			shuffComp = null;
			subMatrix=null;
			matrix=null;
			permutedMatrix = null;
			Runtime.getRuntime().gc();
		}

		randomScores = floatListToArray(randScores);
		
//		Random rand1 = new Random();
//		int counter = 0;
//		System.out.print("rand\t");
//		for (int i = 0; i < randomScores.length; i++) {
//			if (counter > 10000) {
//				break;
//			}
//			int r = rand1.nextInt(randomScores.length);
//			System.out.print(randomScores[r] + ",");
//			counter++;
//		}
//		System.out.println();
//		counter = 0;
//		System.out.print("real\t");
//		for (int i = 0; i < realScores.length; i++) {
//			if (counter > 10000) {
//				break;
//			}
//			int r = rand1.nextInt(realScores.length);
//			System.out.print(realScores[r] + ",");
//			counter++;
//		}
//		System.out.println();
//		realScores = null;
//
//		System.exit(0);

		double randAvg = staticMathHelperMethods.Average(randomScores);
		double randSD = staticMathHelperMethods.sStatistic(randomScores,(float) randAvg);
		
		randAvg = (double)Math.round(randAvg * 100) / 100.0;
		randSD = (double)Math.round(randSD * 100) / 100.0;
		
		NormalDistribution realMates = new NormalDistribution(realAvg, realSD);
		NormalDistribution RandDist = new NormalDistribution(randAvg, randSD);

		// calculate pMates
		double negThreshold = randAvg + K * randSD;
//		negThreshold = randAvg + 1 * randSD;
		double z = (negThreshold - realAvg) / realSD;
		double pMates = 2 * TailProbs.probNormal(z);

//		System.out.println("real mates: " + realMates);
//		System.out.println("random : " + RandDist);
//		System.out.println("p mates: " + pMates);

		llrAnalysisForTscores = new LLRScoresCalculator(
				tScoresCalc.data.getGenesID(), realTscoresGraph);
		llrAnalysisForTscores.setDistributions(realMates, RandDist, pMates);

//		System.out.println("likelihood for extreme positive: "
//				+ llrAnalysisForTscores.getSimLikeLihood(2, true));
//		System.out.println("likelihood for extreme negative: "
//				+ llrAnalysisForTscores.getSimLikeLihood(-2, true));

		dGGraph = llrAnalysisForTscores.getClickWeightedMatrix(false);
		uGGraph = new double[dGGraph.length][dGGraph.length];

		// set the graph as "down"
		// if the diff score if > 0 , set the likelihood as well
		int countPositives = 0;
		int countNegatives = 0;
		for (int i = 0; i < dGGraph.length; i++) {
			for (int j = 0; j < i; j++) {
				if (dGGraph[i][j] > 0 && realTscoresGraph[i][j] > 0) {
					countPositives++;
				}
				if (dGGraph[i][j] > 0 && realTscoresGraph[i][j] < 0) {
					countNegatives++;
				}

				if (realTscoresGraph[i][j] == 0 || dGGraph[i][j] < 0) {
					uGGraph[i][j] = dGGraph[i][j];
				} else {
					if (realTscoresGraph[i][j] > 0) {
						uGGraph[i][j] = dGGraph[i][j];
						dGGraph[i][j] *= -1;
					} else {
						uGGraph[i][j] = -dGGraph[i][j];
					}
				}

				uGGraph[j][i] = uGGraph[i][j];
				dGGraph[j][i] = dGGraph[i][j];
			}
		}
	}


	public void getConsistentCorrelationsGraph() {
		System.out.println("calculating global graph");
		for (int i = 0; i < tScoresCalc.ClassesNames.length; i++) {
			globalCorrelationsLLRAnalyzer = new LLRScoresCalculator(
					tScoresCalc.data.getGenesID(),
					tScoresCalc.class2Correlations.get(i));
			if (i == 0) {
				consistentCorrelationsGraph = globalCorrelationsLLRAnalyzer
						.getClickWeightedMatrix(true);
			} else {
				double[][] tmpGraph = globalCorrelationsLLRAnalyzer
						.getClickWeightedMatrix(true);
				for (int j = 0; j < tmpGraph.length; j++) {
					for (int j2 = 0; j2 < tmpGraph.length; j2++) {
						consistentCorrelationsGraph[j][j2] = Math.min(consistentCorrelationsGraph[j][j2],
								tmpGraph[j][j2]);
					}
				}
			}
		}
	}
	
	public static float[] floatListToArray(LinkedList<Float> positiveScores) {
		float[] posScores = new float[positiveScores.size()];
		int counter = 0;
		for (Iterator<Float> iterator = positiveScores.iterator(); iterator
				.hasNext();) {
			Float score = (Float) iterator.next();
			posScores[counter] = score;
			counter++;
		}
		return posScores;
	}
	
	public void setK(double k) {
		K = k;
	}
}
