package math_utils;

import java.util.Arrays;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Random;



public class staticMathHelperMethods {

	public static float pearsonCorrelation(float[] a, float[] b) {
		float aAvg = Average(a);
		float bAvg = Average(b);
		float aS = sStatistic(a, aAvg);
		float bS = sStatistic(b, bAvg);
		
		if (aS==0 || bS==0){
			return 1;
		}

		float sum = 0;
		for (int i = 0; i < b.length; i++) {
			sum += (a[i] - aAvg) * (b[i] - bAvg);
		}

		float size = a.length - 1;

		return (sum / (size * aS * bS));
	}
	
	public static double getTtestPval(float[] arr,int from,int to){
		
		float avg1 = Average(arr, from, to);
		float[] arr2 = new float[arr.length - (to-from+1)];
		int counter = 0;
		for (int i = 0; i < arr.length; i++) {
			if (i>=from && i<=to){
				continue;
			}
			arr2[counter] = arr[i];
			counter++;
		}
		float avg2 = Average(arr2);
		
		float s1 = variance(arr, avg1,from,to);
		float s2 = variance(arr2, avg2);
		
		if (s1==0 || s2==0){
			return 1;
		}
		
		double df = arr.length-2;
		int NumberOfSicks=to-from+1;
		int NumberOfHealthies=arr.length-NumberOfSicks;
		double svar = ((NumberOfHealthies-1)*s2 + (NumberOfSicks-1)*s1)/df;
		double tScore = (avg1-avg2)/Math.sqrt(svar *((1.0/NumberOfHealthies)+(1.0/NumberOfSicks)));
		return HelperFuncs.betai(0.5*df, 0.5, df/(df + (tScore*tScore)));
		
	}
	
	public static double getTtestPval(double[] arr1,double[] arr2){
		
		double avg1 = Average(arr1);
		double avg2 = Average(arr2);
		
		double s1 = variance(arr1, avg1);
		double s2 = variance(arr2, avg2);
		
		if (s1==0 || s2==0){
			return 1;
		}
		
		double df = arr1.length+arr2.length-2;
		int NumberOfSicks = arr1.length;
		int NumberOfHealthies=arr2.length;
		double svar = ((NumberOfHealthies-1)*s2 + (NumberOfSicks-1)*s1)/df;
		double tScore = (avg1-avg2)/Math.sqrt(svar *((1.0/NumberOfHealthies)+(1.0/NumberOfSicks)));
		return HelperFuncs.betai(0.5*df, 0.5, df/(df + (tScore*tScore)));
		
	}

	public static float pearsonCorrelation(float[] a, float[] b, int from,
			int to) {
		float aAvg = Average(a, from, to);
		float bAvg = Average(b, from, to);
		float aS = sStatistic(a, aAvg, from, to);
		float bS = sStatistic(b, bAvg, from, to);
		
		if (aS==0 || bS==0){
			return 0;
		}

		float sum = 0;
		for (int i = from; i <= to; i++) {
			sum += (a[i] - aAvg) * (b[i] - bAvg);
		}

		float size = to - from;

		return (sum / (size * aS * bS));
	}
	
	public static float robustPearson(float[] a, float[] b, int from, int to){
		Random rand = new Random();
		int num = 10;
		int n = to-from+1;
		int bootstrapSize = n/2  ;
		float[] corrs = new float[num];
		for (int i = 0; i < num; i++) {
			float[] first= new float [bootstrapSize];
			float[] second= new float [bootstrapSize];
			for (int j = 0; j < second.length; j++) {
				int currIndex = rand.nextInt(n)+from;
				first[j]=a[currIndex];
				second[j] = b[currIndex];
			}
			corrs[i]= pearsonCorrelation(first, second);
		}
		Arrays.sort(corrs);
		
		return corrs[corrs.length/2];
	}

	public static float SpearmanCorrelation(float[] a, float[] b) {
		
		//Use a copy of the arrays for sorting
		//and rank calculation
		float[] aCopy = new float[a.length];
		float[] bCopy = new float[a.length];
		Hashtable<Float, Integer> aMapping = new Hashtable<Float, Integer>();
		Hashtable<Float, Integer> bMapping = new Hashtable<Float, Integer>();
		for (int i = 0; i < bCopy.length; i++) {
			aCopy[i]=a[i];
			bCopy[i]=b[i];
			aMapping.put(a[i], i);
			bMapping.put(b[i], i);
		}

		Arrays.sort(aCopy);
		Arrays.sort(bCopy);
		
		
		// These arrays will hold the final ranks
		float[] aCopy2 = new float[a.length];
		float[] bCopy2 = new float[a.length];
		for (int i = 0; i < bCopy.length; i++) {
			int currRankPointerAtA = aMapping.get(aCopy[i]);
			aCopy2[currRankPointerAtA]=i+1;
			
			int currRankPointerAtB = bMapping.get(bCopy[i]);
			bCopy2[currRankPointerAtB]=i+1;
		}
		
		return pearsonCorrelation(aCopy2, bCopy2);
		
	}
	
	public static float pearsonCorrelation(float[] a, float[] b, float avgA,
			float avgB, float aS, float bS) {
		
		if (aS==0 || bS==0){
			return 0;
		}

		float sum = 0;
		for (int i = 0; i < b.length; i++) {
			sum += (a[i] - avgA) * (b[i] - avgB);
		}

		float size = a.length - 1;

		return (sum / (size * aS * bS));
	}

	public static float pearsonCorrelation(float[] a, float[] b, float avgA,
			float avgB, float aS, float bS, int from, int to) {
		
		if (aS==0 || bS==0){
			return 0;
		}

		float sum = 0;
		for (int i = from; i <= to; i++) {
			sum += (a[i] - avgA) * (b[i] - avgB);
		}

		float size = to-from+1;
		float corr = (sum / (size * aS * bS));
		if (corr> 1){
			return 1;
		}
		if (corr < -1){
			return -1;
		}

		return corr; 
	}

	public static float Average(float[] a) {
		float sum = 0;
		float size = a.length;
		for (int i = 0; i < a.length; i++) {
			sum += a[i];
		}
		return sum / size;
	}
	
	public static float Average(float[][] a) {
		float sum = 0;
		float size = a.length*(a.length-1)/2;
		for (int i = 0; i < a.length; i++) {
			for (int j = 0; j < i; j++) {
				sum += a[i][j];
			}
			
		}
		return sum / size;
	}
	
	public static double Average(double[] a) {
		float sum = 0;
		float size = a.length;
		for (int i = 0; i < a.length; i++) {
			sum += a[i];
		}
		return sum / size;
	}

	public static float Average(float[] a, int from, int to) {
		float sum = 0;
		float size = to - from + 1;
		for (int i = from; i <= to; i++) {
			sum += a[i];
		}
		return sum / size;
	}

	public static float sStatistic(float[] a, float avg) {
		float sum = 0;
		float size = a.length - 1;
		for (int i = 0; i < a.length; i++) {
			sum +=(a[i] - avg)*(a[i] - avg);
		}

		Double tmp = Math.sqrt(sum / size);
		
		return tmp.floatValue();
	}
	
	public static float sStatistic(float[][] a, float avg) {
		float sum = 0;
		float size = a.length*(a.length-1)/2;
		size -= 1;
		for (int i = 0; i < a.length; i++) {
			for (int j = 0; j < i; j++) {
				sum +=(a[i][j] - avg)*(a[i][j] - avg);
			}
			
		}
		Double tmp = Math.sqrt(sum / size);
		return tmp.floatValue();
	}
	
	public static float variance(float[] a, float avg){
		float sum = 0;
		float size = a.length - 1;
		for (int i = 0; i < a.length; i++) {
			sum +=(a[i] - avg)*(a[i] - avg);
		}
		return (sum / (size-1));
	}
	
	public static double variance(double[] a, double avg){
		float sum = 0;
		float size = a.length - 1;
		for (int i = 0; i < a.length; i++) {
			sum +=(a[i] - avg)*(a[i] - avg);
		}
		return (sum / (size-1));
	}
	
	public static float variance(float[] a, float avg, int from, int to){
		float sum = 0;
		float size = to - from;
		for (int i = from; i <= to; i++) {
			sum +=(a[i] - avg)*(a[i] - avg);
		}

		return (sum / (size-1));
	}

	public static float sStatistic(float[] a, float avg, int from, int to) {
		float sum = 0;
		float size = to - from;
		for (int i = from; i <= to; i++) {
			sum += Math.pow(a[i] - avg, 2);
		}

		Double tmp = Math.sqrt(sum / size);
		;
		return tmp.floatValue();
	}

	public static float euclideanDistance(float[] a, float[] b) {
		float sum = 0;
		for (int i = 0; i < b.length; i++) {
			sum += Math.pow(a[i] - b[i], 2);
		}

		Double tmp = Math.sqrt(sum);
		return tmp.floatValue();
	}
	
	public static float euclideanDistance(float[] a, float[] b,int from,int to) {
		float sum = 0;
		for (int i = from; i <= to; i++) {
			sum += Math.pow(a[i] - b[i], 2);
		}

		Double tmp = Math.sqrt(sum);
		return -1*tmp.floatValue();
	}

	public static float manhattanDistance(float[] a, float[] b) {
		float sum = 0;
		for (int i = 0; i < b.length; i++) {
			sum += Math.abs(a[i] - b[i]);
		}
		return sum;
	}

	public static void printMatrix(float[][] m) {
		for (int i = 0; i < m.length; i++) {
			for (int j = 0; j < m.length; j++) {
				System.out.print(m[i][j] + "\t");
			}
			System.out.println();
		}
	}

	public static int binarySearch(float[] arr, float val) {
		int first = 0;
		int last = arr.length - 1;

		while (last - first > 1) {
			int middle = (first+last)/2;
			if (arr[middle] > val) {
				last = middle;
			} else {
				first = middle;
			}
		}
		
		if (arr[last]==val){
			return (last);
		}

		return (first);
	}

	public static float[][] computePairWiseScoreMatrix(
			float[][] expressionMatrix, String type) {

		float[] avgs = new float[expressionMatrix.length];
		float[] stdev = new float[expressionMatrix.length];

		for (int i = 0; i < stdev.length; i++) {
			avgs[i] = Average(expressionMatrix[i]);
			stdev[i] = sStatistic(expressionMatrix[i], avgs[i]);
		}

		int n = expressionMatrix.length;
		float[][] pairWiseScoresMatrix = new float[n][n];
		for (int i = 0; i < n; i++) {
			for (int j = 0; j < i; j++) {
				if (type.equalsIgnoreCase("pearson")) {
					pairWiseScoresMatrix[i][j] = pearsonCorrelation(
							expressionMatrix[i], expressionMatrix[j], avgs[i],
							avgs[j], stdev[i], stdev[j]);
				}
				if (type.equalsIgnoreCase("spearman")) {
					pairWiseScoresMatrix[i][j] = SpearmanCorrelation(
							expressionMatrix[i], expressionMatrix[j]);
				}
				if (type.equalsIgnoreCase("euclidean")) {
					pairWiseScoresMatrix[i][j] = euclideanDistance(
							expressionMatrix[i], expressionMatrix[j]);
				}
				if (type.equalsIgnoreCase("manhattan")) {
					pairWiseScoresMatrix[i][j] = manhattanDistance(
							expressionMatrix[i], expressionMatrix[j]);
				}

			}
		}

		for (int i = 0; i < n; i++) {
			for (int j = 0; j < i; j++) {
				pairWiseScoresMatrix[j][i] = pairWiseScoresMatrix[i][j];
			}
		}

		// generally should be one, we will not
		// use these cells in our computation
		// because in our graphs there are no self edges.
		for (int i = 0; i < pairWiseScoresMatrix.length; i++) {
			if (type.equalsIgnoreCase("pearson")) {
				pairWiseScoresMatrix[i][i] = 1;
			}
			if (type.equalsIgnoreCase("euclidean")) {
				pairWiseScoresMatrix[i][i] = 0;
			}
			if (type.equalsIgnoreCase("manhattan")) {
				pairWiseScoresMatrix[i][i] = 0;
			}

		}

		return pairWiseScoresMatrix;
	}
	
	public static float[][] computePairWiseScoreMatrix(
			float[][] expressionMatrix, String type,int from,int to) {

		float[] avgs = new float[expressionMatrix.length];
		float[] stdev = new float[expressionMatrix.length];

		for (int i = 0; i < stdev.length; i++) {
			avgs[i] = Average(expressionMatrix[i],from,to);
			stdev[i] = sStatistic(expressionMatrix[i], avgs[i],from,to);
		}

		int n = expressionMatrix.length;
		float[][] pairWiseScoresMatrix = new float[n][n];
		for (int i = 0; i < n; i++) {
			for (int j = 0; j < i; j++) {
				if (type.equalsIgnoreCase("pearson")) {
					pairWiseScoresMatrix[i][j] = pearsonCorrelation(
							expressionMatrix[i], expressionMatrix[j], avgs[i],
							avgs[j], stdev[i], stdev[j], from, to);
				} else {
					if (type.equalsIgnoreCase("euclidean")) {
						pairWiseScoresMatrix[i][j] = euclideanDistance(
								expressionMatrix[i], expressionMatrix[j],from,to);
					}
					if (type.equalsIgnoreCase("robustPearson")) {
						pairWiseScoresMatrix[i][j] = robustPearson(
								expressionMatrix[i], expressionMatrix[j],from,to);
					}
					if (type.equalsIgnoreCase("manhattan")) {
						pairWiseScoresMatrix[i][j] = manhattanDistance(
								expressionMatrix[i], expressionMatrix[j]);
					}
					if (type.equalsIgnoreCase("spearman")) {
						pairWiseScoresMatrix[i][j] = SpearmanCorrelation(
								expressionMatrix[i], expressionMatrix[j],from,to);
					}
				}
			}
		}

		for (int i = 0; i < n; i++) {
			for (int j = 0; j < i; j++) {
				pairWiseScoresMatrix[j][i] = pairWiseScoresMatrix[i][j];
			}
		}

		// generally should be one, we will not
		// use these cells in our computation
		// because in our graphs there are no self edges.
		for (int i = 0; i < pairWiseScoresMatrix.length; i++) {
			if (type.equalsIgnoreCase("pearson")) {
				pairWiseScoresMatrix[i][i] = 1;
			}
			else{
				if (type.equalsIgnoreCase("euclidean")) {
					pairWiseScoresMatrix[i][i] = 0;
				}
				if (type.equalsIgnoreCase("manhattan")) {
					pairWiseScoresMatrix[i][i] = 0;
				}
			}
		}

		return pairWiseScoresMatrix;
	}

	public static float[] computeSortedScoresArray(float[][] correlationMatrix) {
		int n = correlationMatrix.length;
		float[] correlationArray = new float[n * (n - 1) / 2];
		int counter = 0;
		for (int i = 0; i < n; i++) {
			for (int j = 0; j < i; j++) {
				correlationArray[counter] = correlationMatrix[i][j];
				counter++;
			}
		}

		Arrays.sort(correlationArray);

		return correlationArray;
	}

	public static void copyTwoArrays(float[][] matrix, int from, int to,
			int i, int j, float[] iCopy, float[] jCopy) {
		int locationInCopy = 0;
		for (int k = from; k <= to; k++) {
			jCopy[locationInCopy] = matrix[j][k];
			iCopy[locationInCopy] = matrix[i][k];
			locationInCopy++;
		}
	}

	public static void shuffleArrays(Random rand, float[] iCopy, float[] jCopy) {
		int position = jCopy.length-1;
		for (int l = 0; l < jCopy.length; l++) {
			if (position==0){
				break;
			}
			int iPosition = rand.nextInt(position);
			int jPosition = rand.nextInt(position);

			float tmp = iCopy[iPosition];
			iCopy[iPosition] = iCopy[position - 1];
			iCopy[position - 1] = tmp;

			tmp = jCopy[jPosition];
			jCopy[jPosition] = jCopy[position - 1];
			jCopy[position - 1] = tmp;

			position--;
		}
	}
	
	public static void shuffleArray(Random rand, int[] arr) {
		int position = arr.length-1;
		for (int l = 0; l < arr.length; l++) {
			if (position==0){
				break;
			}
			int iPosition = rand.nextInt(position);

			int tmp = arr[iPosition];
			arr[iPosition] = arr[position - 1];
			arr[position - 1] = tmp;
			position--;
		}
	}
	
	public static float[] getSortedRealScoresSample (float[][] matrix,
			float iterations, String type,int from,int to,boolean dataMatrix) {
		Random rand = new Random();
		int n = matrix.length;
		float[] scoresSample = new float[(new Float(iterations)).intValue()];
		for (int counter = 0; counter < iterations; counter++) {

			int i = rand.nextInt(n);
			int j = rand.nextInt(n);
			while (i==j){
				j = rand.nextInt(n);
			}
			if (dataMatrix){
				if (type.equalsIgnoreCase("pearson")) {
					float corr = pearsonCorrelation(matrix[i], matrix[j],from,to);
					scoresSample[counter] = corr; 
				}
				if (type.equalsIgnoreCase("robustPearson")) {
					scoresSample[counter] = robustPearson(matrix[i],matrix[j],from,to);
				}
				if (type.equalsIgnoreCase("euclidean")) {
					scoresSample[counter] = euclideanDistance(matrix[i], matrix[j],from,to);
				}
				if (type.equalsIgnoreCase("manhattan")) {
					scoresSample[counter] = manhattanDistance(matrix[i], matrix[j]);
				}
				if (type.equalsIgnoreCase("spearman")) {
					scoresSample[counter] = SpearmanCorrelation(matrix[i], matrix[j], from, to);
				}

			}
			else{
				scoresSample[counter] = matrix[i][j];
			}
			
		}

		Arrays.sort(scoresSample);
		return scoresSample;
	}

	public static float SpearmanCorrelation(float[] a,float[] b, int from, int to) {
		float[] iCopy = new float[to-from+1];
		float[] jCopy = new float[to-from+1];
		int counter = 0;
		for (int i = from; i <= to; i++) {
			iCopy[counter] = a[i];
			jCopy[counter] = b[i];
			counter++;
		}
		return SpearmanCorrelation(iCopy, jCopy);
	}
	
	public static float[] getSortedRealScoresSample (float[][] matrix,
			float iterations, String type,boolean dataMatrix) {
		Random rand = new Random();
		int n = matrix.length;
		float[] scoresSample = new float[(new Float(iterations)).intValue()];
		for (int counter = 0; counter < iterations; counter++) {

			int i = rand.nextInt(n);
			int j = rand.nextInt(n);
			while (i==j){
				j = rand.nextInt(n);
			}
			if (dataMatrix){
				if (type.equalsIgnoreCase("pearson")) {
					float corr = pearsonCorrelation(matrix[i], matrix[j]);
					scoresSample[counter] = corr; 
				}
				if (type.equalsIgnoreCase("euclidean")) {
					scoresSample[counter] = euclideanDistance(matrix[i], matrix[j]);
				}
				if (type.equalsIgnoreCase("manhattan")) {
					scoresSample[counter] = manhattanDistance(matrix[i], matrix[j]);
				}
				if (type.equalsIgnoreCase("spearman")) {
					scoresSample[counter] = SpearmanCorrelation(matrix[i], matrix[j]);
				}
				

			}
			else{
				scoresSample[counter] = matrix[i][j];
			}
			
		}

		Arrays.sort(scoresSample);
		return scoresSample;
	}

	public static void standardizeByGenes(float[][] matrix,int[] borders) {
		
		for (int i = 0; i < borders.length; i+=2) {
			int size = borders[i+1]-borders[i]+1;
			for (int k = 0; k < matrix.length; k++) {
				float geneAvg = 0;
				for (int j = borders[i]; j <= borders[i+1]; j++) {
					geneAvg += matrix[k][j];
				}
				geneAvg /= (float)size;
				float geneSdv = 0;
				for (int j = borders[i]; j <= borders[i+1]; j++) {
					float tmp = matrix[k][j] - geneAvg;
					geneSdv += tmp * tmp;
				}
				geneSdv /= (size - 1);
				Double tmpS = Math.sqrt(geneSdv);
				for (int j = borders[i]; j <= borders[i+1]; j++) {
					if (geneSdv ==0.0){
						matrix[k][j] = 0;
					}
					else{
						matrix[k][j] = (matrix[k][j] - geneAvg) / tmpS.floatValue();
					}
					
				}

			}
		}
		
	}
	
	public static void standardizeByGenes(float[][] matrix) {

		int size = matrix[0].length;
		for (int k = 0; k < matrix.length; k++) {
			float geneAvg = 0;
			for (int j = 0; j < size; j++) {
				geneAvg += matrix[k][j];
			}
			geneAvg /= size;
			float geneSdv = 0;
			for (int j = 0; j < size; j++) {
				float tmp = matrix[k][j] - geneAvg;
				geneSdv += tmp * tmp;
			}
			geneSdv /= (size - 1);
			Double tmpS = Math.sqrt(geneSdv);
			for (int j = 0; j < size; j++) {
				if (geneSdv == 0) {
					matrix[k][j] = 0;
				} else {
					matrix[k][j] = (matrix[k][j] - geneAvg) / tmpS.floatValue();
				}

			}

		}

	}

	public static float[][] getMatrixDeepClone(float[][] mat) {
		int n = mat.length;
		float[][] clone = new float[n][n];
		for (int i = 0; i < clone.length; i++) {
			for (int j = 0; j < clone.length; j++) {
				clone[i][j] = mat[i][j];
			}
		}
		return clone;
	}
	
	public static float estimateFDRThreshold(float[][] dataMatrix, float fdr,float[] realArr,String type
			,int from,int to) {
		int realArrayPosition = 0;
		int fakeArrayPosition = 0;
		
		Random rand = new Random();
		
		float[] fakeArr = getShuffeledCorrelationEstimates(dataMatrix,type,realArr.length, rand,from,to);

		float acceptedNumber = realArr.length;
		float fakeCurrentNumber = fakeArr.length ;
		float currentFdr = 0;
		
		while (realArrayPosition < realArr.length 
				&& realArr[realArrayPosition]< (float) 0.1){
			realArrayPosition++;
		}

		while (realArrayPosition<realArr.length) {
			
			while (fakeArrayPosition < fakeArr.length
					&& realArr[realArrayPosition] > fakeArr[fakeArrayPosition]) {
				fakeArrayPosition++;
			}

			fakeCurrentNumber = fakeArr.length - fakeArrayPosition;
			currentFdr = (fakeCurrentNumber / (float) fakeArr.length)
					/ (acceptedNumber / (float) realArr.length);
			

			if (currentFdr <= fdr){
				return realArr[realArrayPosition];
			}
			
			realArrayPosition++;
			acceptedNumber--;
			
		}
		
		// In case no fdr was found return the p quantile
		int quantilePosition = (new Float(fdr*realArr.length)).intValue();
		return realArr[quantilePosition];
	}
	
	public static float[] getShuffeledCorrelationEstimates(float[][] matrix, String type,int iterations,Random rand,int from,int to){
		
		float[] arr = new float[iterations];
		
		for (int i = 0; i < arr.length; i++) {
			float[] a = new float[to-from + 1];
			float[] b = new float[to-from + 1];
			
			int index1 = rand.nextInt(matrix.length);
			int index2 = rand.nextInt(matrix.length);
			
			int counter = 0;
			for (int j = from; j <= to; j++) {
				a[counter] = matrix[index1][j];
				b[counter] = matrix[index2][j];
				counter++;
			}
			
			shuffleArrays(rand, a, b);
			if (type.equalsIgnoreCase("pearson")){
				arr[i] = pearsonCorrelation(a, b);
			}
			if (type.equalsIgnoreCase("spearman")){
				arr[i]=SpearmanCorrelation(a, b);
			}
			
 		}
		
		Arrays.sort(arr);
		
		return arr;
		
	}
	
	public static double binomialCoefficient(int n,int k){
		double result =1;
		for (int i = 1; i <= k; i++) {
			result *= ((double)(n-(k-i))/(double)i);
		}
		return result;
	}
	
	public static double getHyperGeometricPval(int successes,int sampleSize,int successesInPopulation,int populationSize){
		double pval = 0;
		double numPossibleSelections = binomialCoefficient(populationSize, sampleSize);
		for (int i = successes; i <= sampleSize; i++) {
			if (i > successesInPopulation){
				break;
			}
			double possibleSuccessesSelections = binomialCoefficient(successesInPopulation, i);
			System.out.println(successesInPopulation + " choose " + i);
			System.out.println(possibleSuccessesSelections);
			double possibleFailures = binomialCoefficient(populationSize-successesInPopulation, sampleSize-i);
			System.out.println(possibleFailures);
			System.out.println(i + " :" + (possibleFailures*possibleSuccessesSelections)/numPossibleSelections);
			pval += (possibleFailures*possibleSuccessesSelections)/numPossibleSelections;
		}
		
		return pval;
	}
	
	public static double HGmean(int sampleSize,int successesInPopulation,int populationSize){
		double result = (double)sampleSize*(double)successesInPopulation;
		result /= (double) populationSize;
		return result;
	}
	
	public static double HGSD (int sampleSize,int successesInPopulation,int populationSize){
		double a = (double)sampleSize*(double)successesInPopulation * (double)(populationSize-sampleSize)*(double)(populationSize-successesInPopulation);
		if (a<0){
			System.out.println(sampleSize);
			System.out.println(successesInPopulation);
			System.out.println(populationSize);
			System.out.println("Error in HG variance settings " + a);
			System.exit(1);
		}
		double b = (double)populationSize*(double)populationSize*(double)(populationSize-1);
		double var = a/b;
		return Math.sqrt(var);
	}
	
	public static double getHyperGeometricPvalEstimates(int successes,int sampleSize,int successesInPopulation,int populationSize){
		
		double mean = HGmean(sampleSize, successesInPopulation, populationSize);
		double sd = HGSD(sampleSize, successesInPopulation, populationSize);
		
		double numSDDevs = (double)successes - mean;
		numSDDevs /= sd;
		if (numSDDevs <= 0){
			return 1;
		}
		else{
			return (1.0/(1.0 + numSDDevs*numSDDevs));
		}
		

	}
	
	public static double getCDFofBinomialDist(int successes,int trails,float p){
		double result = 0;
		for (int i = 0; i <= successes; i++) {
			double pToI = Math.pow(p, i);
			double compPToI = Math.pow(1-p, trails-i);
			double probCoeff = pToI*compPToI;
			result += binomialCoefficient(trails, i)*probCoeff;
			
		}
		return result;
	}
	
	public static float ANOVAstatistic(float[] vals, int[] borders,Map<Integer,Integer> index2class,int[] sizes){
		
		float k = borders.length/2;
		float[] classesSums = new float[borders.length/2];
		float[] normalizedSquaredSums = new float[borders.length/2];
		
		float totalSum = 0;
		float totalInstances = vals.length;
		for (int i = 0; i < vals.length; i++) {
			int currClass = index2class.get(i);
			classesSums[currClass]+=vals[i];
			totalSum+=vals[i];
		}
		
		float SSbg = 0;
		for (int i = 0; i < k; i++) {
			normalizedSquaredSums[i]= (classesSums[i]*classesSums[i]/(float) sizes[i]);
			SSbg+= normalizedSquaredSums[i];
		}
		
		SSbg -= (totalSum*totalSum/totalInstances);
		
		float SSwg = 0;
		for (int i = 0; i < k; i++) {
			float currMoment2 = 0;
			for (int j = borders[2*i]; j <=borders[2*i+1] ; j++) {
				currMoment2 += (vals[j]*vals[j]);
			}
			SSwg += (currMoment2 - normalizedSquaredSums[i]);
		}
		
		float MSbg = SSbg/(float)(k-1);
		float MSwg = SSwg / (totalInstances-k);
		
		return (MSbg/MSwg);
	}
	
	
	public static float getMaximalSupervisedSimilairtyTScore(float[] a,float[] b,int[] borders,int[] cSizes,Random rand){
		Hashtable<Integer, Float> cl2Vars = new Hashtable<Integer, Float>();
		Hashtable<Integer, Float> cl2Avgs = new Hashtable<Integer, Float>();
		Hashtable<Integer, Float> cl2SD = new Hashtable<Integer, Float>();
		
		int numVals = 100;
		
		// Go over all classes
		for (int i = 0; i < cSizes.length; i++) {
			
			// for each class get its average and variance of 
			// distance between points
			int from = borders[2*i];
			int to = borders[2*i+1];
			double sum=0;
			LinkedList<Double> distances = new LinkedList<Double>();
			
			for (int j = 0; j < numVals; j++) {
				int firstPointIndex = rand.nextInt(cSizes[i])+from;
				int secondPointIndex = rand.nextInt(cSizes[i]-1)+from;
				if (firstPointIndex==secondPointIndex){
					secondPointIndex = to;
				}
				double dist =0;
				double tmpVal = a[firstPointIndex]-a[secondPointIndex];
				dist += tmpVal*tmpVal;
				tmpVal = b[firstPointIndex]-b[secondPointIndex];
				dist += tmpVal*tmpVal;
				sum+=dist;
				distances.add(dist);
				
			}
			
			double avg = sum / (double) numVals;
			double variance = 0;
			for (Iterator<Double> iterator = distances.iterator(); iterator.hasNext();) {
				Double value = (Double) iterator.next();
				double tmpVal = value-avg;
				variance += tmpVal*tmpVal;
			}
			
			variance /= (double) numVals;
			cl2Avgs.put(i, (new Double(avg)).floatValue());
			cl2SD.put(i, (new Double(Math.sqrt(variance))).floatValue());
			cl2Vars.put(i, (new Double(variance)).floatValue());
		}
		
		// Get pairwise maximal T statistic
		float maxAbsScore = 0;
		for (int i = 0; i < cSizes.length; i++) {
			for (int j = 0; j < i; j++) {
				float sharedVar = (cl2Vars.get(i)/(float)numVals)+(cl2Vars.get(j)/(float)numVals);
				float tScore = (cl2Avgs.get(i)-cl2Avgs.get(j))/sharedVar;
				if (tScore < 0){
					tScore *=-1;
				}
				if (tScore > maxAbsScore){
					maxAbsScore=tScore;
				}
			}
		}
		
		return maxAbsScore;
	}
	
	public static float getMaximalSupervisedSimilairtyROCScore(float[] a,float[] b,int[] borders,int[] cSizes,Random rand){
		Hashtable<Integer, float[]> cl2Vals = new Hashtable<Integer, float[]>();
		
		int numVals = 100;
		
		// Go over all classes
		for (int i = 0; i < cSizes.length; i++) {
			
			cl2Vals.put(i, new float[numVals]);
			// for each class get its average and variance of 
			// distance between points
			int from = borders[2*i];
			int to = borders[2*i+1];
			for (int j = 0; j < numVals; j++) {
				int firstPointIndex = rand.nextInt(cSizes[i])+from;
				int secondPointIndex = rand.nextInt(cSizes[i]-1)+from;
				if (firstPointIndex==secondPointIndex){
					secondPointIndex = to;
				}
				double dist =0;
				double tmpVal = a[firstPointIndex]-a[secondPointIndex];
				dist += tmpVal*tmpVal;
				tmpVal = b[firstPointIndex]-b[secondPointIndex];
				dist += tmpVal*tmpVal;
				cl2Vals.get(i)[j]=(new Double(dist)).floatValue();
			}
			
			Arrays.sort(cl2Vals.get(i));
		}
		
		// Get pairwise maximal ROC statistic
		float maxAbsScore = 0;
		for (int i = 0; i < cSizes.length; i++) {
			for (int j = 0; j < i; j++) {
				float roc = calcROCscore(cl2Vals.get(i), cl2Vals.get(j));
				if (roc > maxAbsScore){
					maxAbsScore = roc;
				}
			}
		}
		
		return maxAbsScore;
	}
	
	
	public static float calcROCscore(float[] a,float[] b){
		
		float FPR = 0;
		
		float numPos = a.length;
		float numNeg = b.length;
		
		int posPointer = a.length-1;
		int negPointer = b.length-1;
		
		float roc = 0;
		while (posPointer > 0 || negPointer > 0){
			float lastFPR = FPR;
			
			if (posPointer == 0){
				negPointer--;
			}
			else if (negPointer == 0){
				posPointer--;
			}
			else if (a[posPointer] < b[negPointer] ){
				negPointer--;
			}
			else{
				posPointer--;
			}
			
			float FN = posPointer+1;
			float TN = negPointer+1;
			float FP = numNeg - TN;
			float TP = numPos - FN;
			
			float newTPR = TP / numPos;
			float newFPR = FP / numNeg;
			
//			System.out.println("new TPR: " + newTPR);
//			System.out.println("new FPR: " + newFPR);
			
			float xDiff = newFPR-lastFPR;
			roc+= xDiff*newTPR;
			
			FPR = newFPR;

		}
		
		if (roc < 0.5){
			roc = 1-roc;
		}
//		System.out.println(roc);
		return roc;
	}

}
