package scripts;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;

import main_scripts.DICER_main;
import math_utils.ExpressionMatrix;

/**
 * This script calculates DICER's stability. DICER is a randomized algorithm.
 * 
 */

public class CalcDICERStability {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		if (args.length > 5 || args.length < 3) {
			System.out
					.println("usage:<GE matrix><classes file><classIndex><pathways mapping path>"
							+ "optional: <other sol>");
			System.exit(0);
		}

		ExpressionMatrix data = new ExpressionMatrix(args[0]);
		String classesFilePath = args[1];
		int classIndex = Integer.parseInt(args[2]);

		Hashtable<String, HashSet<String>> priorSol = new Hashtable<String, HashSet<String>>();
		Hashtable<String, HashSet<Integer>> priorSolAsIndex = new Hashtable<String, HashSet<Integer>>();
		LinkedList<HashSet<Integer>> pkIndicesAsList = new LinkedList<HashSet<Integer>>();
		if (args.length >= 4) {
			try {
				BufferedReader reader = new BufferedReader(new FileReader(
						args[4]));
				String l = "";

				while ((l = reader.readLine()) != null) {
					String[] arr = l.split("\t");
					String gene = arr[0];
					String set = arr[1];

					if (!priorSol.containsKey(set)) {
						priorSol.put(set, new HashSet<String>());
						priorSolAsIndex.put(set, new HashSet<Integer>());
					}

					priorSol.get(set).add(gene);
					priorSolAsIndex.get(set).add(data.genesID.indexOf(gene));
				}

				reader.close();

			} catch (FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

			for (Iterator<String> iterator = priorSol.keySet().iterator(); iterator
					.hasNext();) {
				String setName = (String) iterator.next();
				pkIndicesAsList.add(priorSolAsIndex.get(setName));
			}
		}

		Hashtable<Integer, Hashtable<String, HashSet<String>>> repeat2module1 = new Hashtable<Integer, Hashtable<String, HashSet<String>>>();
		Hashtable<Integer, HashSet<String>> repeat2enrichements = new Hashtable<Integer, HashSet<String>>();
		Hashtable<Integer, boolean[][]> repeat2MetaGraph = new Hashtable<Integer, boolean[][]>();
		Hashtable<Integer, boolean[][]> repeat2PriorSolMetaGraph = new Hashtable<Integer, boolean[][]>();

		int repeats = 10;
		for (int i = 0; i < repeats; i++) {
			DICER_main dicer = new DICER_main(data, classesFilePath, classIndex);
			dicer.runAlgo(true, false);
			repeat2module1.put(i, dicer.getModules());
			HyperGeometricEnrichmentAnalysis anal = new HyperGeometricEnrichmentAnalysis(
					args[3], data.getGenesID(), dicer.getModules());
			anal.calcResults(0.05);
			repeat2enrichements.put(i, anal.getEnrichedTerms());
			anal.printResults(0.05);
			anal.printResultsStatistics(0.05);
			repeat2MetaGraph.put(i, dicer.getMetaGraph());
			repeat2PriorSolMetaGraph
					.put(i, dicer.getMetaGraph(pkIndicesAsList));
		}

		double avgGenesJ = 0;
		double avgEnrichmentsJ = 0;
		double matricesJ = 0;
		if (args.length >= 4) {

			HyperGeometricEnrichmentAnalysis pkAnal = new HyperGeometricEnrichmentAnalysis(
					args[3], data.genesID, priorSol);
			System.out.println("given solution enrichment stats");
			pkAnal.printResultsStatistics(0.05);
			pkAnal.calcResults(0.05);
			pkAnal.printResults(0.05);
			for (int i = 0; i < repeats; i++) {
				double j = avgRunningMax(priorSol, repeat2module1.get(i));
				avgGenesJ += j;
				double enrichmentJ = getJaccardScore(pkAnal.getEnrichedTerms(),
						repeat2enrichements.get(i));
				avgEnrichmentsJ += enrichmentJ;
				matricesJ += getJaccardScore(repeat2MetaGraph.get(i),
						repeat2PriorSolMetaGraph.get(i));
			}
			avgGenesJ /= repeats;
			avgEnrichmentsJ /= repeats;
			matricesJ /= repeats;
		}

		double withinGeneJ = 0;
		double withinEnrhichmentJ = 0;
		double withinMatricesJ = 0;
		double c = 0;
		for (int i = 0; i < repeats; i++) {
			for (int j = 0; j < i; j++) {
				double iVsj = avgRunningMax(repeat2module1.get(i),
						repeat2module1.get(j));
				double jVsi = avgRunningMax(repeat2module1.get(j),
						repeat2module1.get(i));
				double currJ = (iVsj + jVsi) / 2.0;
				withinGeneJ += currJ;

				double enrichmentJ = getJaccardScore(
						repeat2enrichements.get(i), repeat2enrichements.get(j));
				withinEnrhichmentJ += enrichmentJ;

				withinMatricesJ += getJaccardScore(repeat2MetaGraph.get(i),
						repeat2MetaGraph.get(j));

				c++;
			}
		}
		withinGeneJ /= c;
		withinEnrhichmentJ /= c;
		withinMatricesJ /= c;

		System.out
				.println("##########################################################################################");
		System.out
				.println("############################ finished running dicer repeats ##############################");
		System.out
				.println("avergae gene sets Jaccard vs. the known solution:\t"
						+ avgGenesJ);
		System.out.println("avergae gene sets Jaccard of repeats:\t"
				+ withinGeneJ);
		System.out.println("average enrichments jaccard vs. known:\t"
				+ avgEnrichmentsJ);
		System.out.println("average enrichments jaccard of repeats:\t"
				+ withinEnrhichmentJ);
		System.out
				.println("Jaccard of Pair membership in meta-module, vs. known:\t"
						+ matricesJ);
		System.out
				.println("Jaccard of Pair membership in meta-module, within:\t"
						+ withinMatricesJ);

	}

	public static double getJaccardScore(Collection<String> A,
			Collection<String> B) {
		Collection<String> union = new HashSet<String>(A);
		union.addAll(B);

		Collection<String> intersect = new HashSet<String>();
		for (Iterator<String> iterator = A.iterator(); iterator.hasNext();) {
			String inA = (String) iterator.next();
			if (B.contains(inA)) {
				intersect.add(inA);
			}
		}

		double j = (double) intersect.size() / (double) union.size();
		return j;
	}

	public static double getJaccardScore(boolean[][] A, boolean[][] B) {

		double union = 0;
		double intersect = 0;

		for (int i = 0; i < B.length; i++) {
			for (int j = 0; j < i; j++) {
				if (A[i][j] || B[i][j]) {
					union++;
				}
				if (A[i][j] && B[i][j]) {
					intersect++;
				}
			}
		}
		
		System.out.println("union size:\t" + union);
		System.out.println("intersection:\t" + intersect);

		return intersect / union;
	}

	public static double avgRunningMax(Hashtable<String, HashSet<String>> A,
			Hashtable<String, HashSet<String>> B) {
		double avgJ = 0;
		for (Iterator<String> iterator = A.keySet().iterator(); iterator
				.hasNext();) {
			String inAName = (String) iterator.next();
			HashSet<String> inASet = A.get(inAName);
			double maxJ = 0;
			for (Iterator<String> iterator2 = B.keySet().iterator(); iterator2
					.hasNext();) {
				String inBName = (String) iterator2.next();
				HashSet<String> inBSet = B.get(inBName);
				double j = getJaccardScore(inASet, inBSet);
				if (j > maxJ) {
					maxJ = j;
				}
			}
			// System.out.println(maxJ);
			avgJ += maxJ;
		}
		avgJ /= (double) A.keySet().size();
		return avgJ;
	}

	public static double weightedRunningMax(
			Hashtable<String, HashSet<String>> A,
			Hashtable<String, HashSet<String>> B) {
		double avgJ = 0;
		double sum = 0;
		for (Iterator<String> iterator = A.keySet().iterator(); iterator
				.hasNext();) {
			String inAName = (String) iterator.next();
			HashSet<String> inASet = A.get(inAName);
			double maxJ = 0;
			for (Iterator<String> iterator2 = B.keySet().iterator(); iterator2
					.hasNext();) {
				String inBName = (String) iterator2.next();
				HashSet<String> inBSet = B.get(inBName);
				double j = getJaccardScore(inASet, inBSet);
				if (j > maxJ) {
					maxJ = j;
				}
			}
			// System.out.println(maxJ);
			sum += inASet.size();
			maxJ *= inASet.size();
			// System.out.println(maxJ);
			avgJ += maxJ;
		}
		avgJ /= sum;
		return avgJ;
	}

}
