package scripts;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;

import scripts.HyperGeometricEnrichmentAnalysis;

import math_utils.ExpressionMatrix;


import edu.tau.compbio.math.TailProbs;

public class HyperGeometricEnrichmentAnalysis {

	// constants
	int minOverlap = 3;
	int maxSetSize = 2000;
	double jaccardThreshold = 1;

	ArrayList<String> bg;
	Hashtable<String, HashSet<String>> clusters = new Hashtable<String, HashSet<String>>();
	Hashtable<String, HashSet<String>> geneSets = new Hashtable<String, HashSet<String>>();
	
	ArrayList<oneEnrichmentResult> allResults = new ArrayList<HyperGeometricEnrichmentAnalysis.oneEnrichmentResult>();
	ArrayList<oneEnrichmentResult> passedResults = new ArrayList<HyperGeometricEnrichmentAnalysis.oneEnrichmentResult>();
	Hashtable<String, ArrayList<oneEnrichmentResult>> cluster2acceptedResults;
	HashSet<String> enrichedClusters = new HashSet<String>();
	
	double afterCorrectionThr = -1;

	public HyperGeometricEnrichmentAnalysis(String geneSetsFile,
			ArrayList<String> bg, Hashtable<String, HashSet<String>> c) {

		clusters = c;
		this.bg = bg;
		BufferedReader reader;
		try {
			reader = new BufferedReader(new FileReader(geneSetsFile));
			String currLine = "";
			while ((currLine = reader.readLine()) != null) {
				String[] parsedLine = currLine.split("\t");
				if (parsedLine.length < 2) {
					continue;
				}
				String setName = parsedLine[1];
				String gene = parsedLine[0];

				if (!bg.contains(gene)) {
					continue;
				}

				if (!geneSets.containsKey(setName)) {
					HashSet<String> set = new HashSet<String>();
					set.add(gene);
					geneSets.put(setName, set);
				} else {
					geneSets.get(setName).add(gene);
				}

			}
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		Hashtable<String, HashSet<String>> reducedGeneSets = new Hashtable<String, HashSet<String>>();
		for (Iterator<String> iterator = geneSets.keySet().iterator(); iterator.hasNext();) {
			String setName = (String) iterator.next();
			if (geneSets.get(setName).size()<maxSetSize){
				reducedGeneSets.put(setName, geneSets.get(setName));
			}
		}
		
		geneSets = reducedGeneSets;

	}

	public void calcResults() {

		allResults = new ArrayList<HyperGeometricEnrichmentAnalysis.oneEnrichmentResult>();
		for (Iterator<String> iterator = clusters.keySet().iterator(); iterator
				.hasNext();) {
			String clusterName = (String) iterator.next();
			HashSet<String> cluster = clusters.get(clusterName);

			for (Iterator<String> iterator2 = geneSets.keySet().iterator(); iterator2
					.hasNext();) {
				String setName = (String) iterator2.next();
				HashSet<String> set = geneSets.get(setName);

				// calc intersection
				int intersect = 0;
				for (Iterator<String> iterator3 = set.iterator(); iterator3
						.hasNext();) {
					String inSet = (String) iterator3.next();
					if (cluster.contains(inSet)) {
						intersect++;
					}
				}
				if (intersect < minOverlap) {
					continue;
				}

				// calc pval
				// System.out.println("bg size: " + bg.size());
				// System.out.println("set name" + setName + ", size is " +
				// set.size());
				// System.out.println("cluster name: " + clusterName +
				// ", cluster size is: " + cluster.size());
				// System.out.println("intersect: " + intersect);
				double pVal = 1 - TailProbs.probHyperGeometricC(bg.size(),
						set.size(), cluster.size(), intersect);
				oneEnrichmentResult r = new oneEnrichmentResult(clusterName,setName);
				
				r.intersect = intersect;
				r.pVal = pVal;
				
				// calc enrichment factor
				float numInCluster = r.commonGenes.size();
				float probInCluster = numInCluster/(float)clusters.get(r.cluster).size();
				
				float numInData = geneSets.get(r.set).size();
				float probInData = numInData/(float)bg.size();
				
				r.enrichmentFactor = probInCluster/probInData;
				
				allResults.add(r);
			}

		}

		Collections.sort(allResults);
	}

	public void determineFDRThrIndependenceAssumption(double p) {
		double m = allResults.size();
		for (int i = allResults.size() - 1; i >= 0; i--) {
			if (allResults.get(i).pVal <= (((double) i / m) * p)) {
				afterCorrectionThr = allResults.get(i).pVal;
				break;
			}
		}
		
		passedResults = new ArrayList<HyperGeometricEnrichmentAnalysis.oneEnrichmentResult>();
		cluster2acceptedResults = new Hashtable<String, ArrayList<oneEnrichmentResult>>();
		
		
		for (int i = allResults.size() - 1; i >= 0; i--) {
			if (allResults.get(i).pVal <= afterCorrectionThr) {
				passedResults.add(allResults.get(i));
				enrichedClusters.add(allResults.get(i).cluster);
				
				if (!cluster2acceptedResults.containsKey(allResults.get(i).cluster)){
					cluster2acceptedResults.put(allResults.get(i).cluster, new ArrayList<HyperGeometricEnrichmentAnalysis.oneEnrichmentResult>());
				}
				cluster2acceptedResults.get(allResults.get(i).cluster).add(allResults.get(i));
			}
		}
		
		// remove duplications using jaccard coeff
		Hashtable<String, ArrayList<oneEnrichmentResult>> clust2reduced = new Hashtable<String, ArrayList<oneEnrichmentResult>>();
		for (Iterator<String> iterator = cluster2acceptedResults.keySet().iterator(); iterator.hasNext();) {
			String cluster = (String) iterator.next();
			
			// this array is ordered in decending order, by significance
			ArrayList<oneEnrichmentResult> accepted = cluster2acceptedResults.get(cluster);
			ArrayList<oneEnrichmentResult> reduced = new ArrayList<HyperGeometricEnrichmentAnalysis.oneEnrichmentResult>();
			for (int i = 0; i < accepted.size(); i++) {
				boolean isdup = false;
				HashSet<String> currSet = accepted.get(i).commonGenes;
				
				for (Iterator<oneEnrichmentResult> iterator2 = reduced.iterator(); iterator2.hasNext();) {
					oneEnrichmentResult inreduced = (oneEnrichmentResult) iterator2.next();
					HashSet<String> currInreduced = inreduced.commonGenes;
					
					// calculate jaccard
					double intersection = 0;
					for (Iterator<String> iterator3 = currInreduced.iterator(); iterator3
							.hasNext();) {
						String a = (String) iterator3.next();
						if (currSet.contains(a)){
							intersection++;
						}
					}
					double union = currSet.size()+currInreduced.size()-intersection;
					double jaccard = intersection/union;
					if (jaccard >= jaccardThreshold){
						isdup = true;
						break;
					}
				}
				
				if (!isdup){
					reduced.add(accepted.get(i));
				}
				
			}
			
			clust2reduced.put(cluster, reduced);
		}
		
		cluster2acceptedResults = clust2reduced;
	}
	

	public void printResults(double p) {

		calcResults();
		determineFDRThrIndependenceAssumption(p);
		for (int i = 0; i < passedResults.size(); i++) {
			System.out.println(passedResults.get(i));
		}
	}
	
	public void calcResults(double p) {
		calcResults();
		determineFDRThrIndependenceAssumption(p);
	}
	
	public HashSet<String> getEnrichedTerms(){
		HashSet<String> names = new HashSet<String>();
		for (Iterator<oneEnrichmentResult> iterator = passedResults.iterator(); iterator.hasNext();) {
			oneEnrichmentResult result = (oneEnrichmentResult) iterator.next();
			names.add(result.set);
		}
		
		return names;
	}
	
	public void printResultsStatistics(double p) {

		calcResults(p);

		int numAccepted = 0;
		HashSet<String> enrichedClusters = new HashSet<String>();
		HashSet<String> enrichedTerms = new HashSet<String>();
		double avgEnrichmentFactor = 0;
		for (int i = 0; i < allResults.size(); i++) {
			if (allResults.get(i).pVal > afterCorrectionThr) {
				break;
			}
			numAccepted++;
			enrichedClusters.add(allResults.get(i).cluster);
			avgEnrichmentFactor+=allResults.get(i).enrichmentFactor;
			enrichedTerms.add(allResults.get(i).set);
		}
		avgEnrichmentFactor/=(double)numAccepted;
		System.out.println(numAccepted+"\t"+enrichedTerms.size()+"\t"+enrichedClusters.size()+"\t"+avgEnrichmentFactor);
	}
	
	public void printResultsStatisticsRemoveDuplications(double p) {

		calcResults();
		determineFDRThrIndependenceAssumption(p);

		int numAccepted = 0;
		HashSet<String> enrichedClusters = new HashSet<String>();
		HashSet<String> enrichedTerms = new HashSet<String>();
		double avgEnrichmentFactor = 0;
		
		for (Iterator<String> iterator = cluster2acceptedResults.keySet().iterator(); iterator.hasNext();) {
			String clust = (String) iterator.next();
			numAccepted += cluster2acceptedResults.get(clust).size();
			ArrayList<oneEnrichmentResult> acc = cluster2acceptedResults.get(clust);
			for (Iterator<oneEnrichmentResult> iterator2 = acc.iterator(); iterator2.hasNext();) {
				oneEnrichmentResult a = (oneEnrichmentResult) iterator2
						.next();
				avgEnrichmentFactor += a.enrichmentFactor;
				
			}
		}
		
		avgEnrichmentFactor/=(double)numAccepted;
		System.out.println(numAccepted+"\t"+enrichedTerms.size()+"\t"+enrichedClusters.size()+"\t"+avgEnrichmentFactor);
	}
	
	public void printGenes() {

		for (int i = 0; i < allResults.size(); i++) {
			if (allResults.get(i).pVal > afterCorrectionThr) {
				break;
			}
			allResults.get(i).printCommonGenes();
		}
	}

	class oneEnrichmentResult implements Comparable<oneEnrichmentResult> {
		String cluster;
		String set;
		HashSet<String> commonGenes;
		int intersect;
		double pVal;
		double enrichmentFactor;
		
		public oneEnrichmentResult(String c,String s) {
			cluster=c;
			set=s;
			
			HashSet<String> setGenes = geneSets.get(set);
			HashSet<String> clusterGenes = clusters.get(cluster);
			commonGenes = new HashSet<String>();
			for (Iterator<String> iterator = clusterGenes.iterator(); iterator
					.hasNext();) {
				String inClust = (String) iterator.next();
				if (setGenes.contains(inClust)){
					commonGenes.add(inClust);
				}
			}
		}
		
		public void printCommonGenes(){
			String name = set+"_"+cluster;
			for (Iterator<String> iterator = commonGenes.iterator(); iterator.hasNext();) {
				String gene = (String) iterator.next();
				System.out.println(gene+"\t"+name);
			}
		}
		
		@Override
		public String toString() {
			// TODO Auto-generated method stub
			return ""+cluster + "\t"
			+ set + "\t" + pVal + "\t"
			+ intersect + "\t" + enrichmentFactor;
			
		}

		@Override
		public int compareTo(oneEnrichmentResult o) {
			if (pVal > o.pVal) {
				return 1;
			} else {
				return -1;
			}
		}
	}

	public static void main(String[] args) {
		if (args.length != 4) {
			System.out
					.println("Usage: <gene sets><clustering solution><expression matrix for backgroumd><fdr>");
			System.exit(1);
		}
		
		ExpressionMatrix mat = new ExpressionMatrix(args[2]);
		HashSet<String> bg = new HashSet<String>(mat.genesID);

		Hashtable<String, HashSet<String>> geneSets = readGeneSets(args[1], bg);
		
		HyperGeometricEnrichmentAnalysis anal = new HyperGeometricEnrichmentAnalysis(
				args[0], mat.getGenesID(), geneSets);
		anal.printResults(Double.parseDouble(args[3]));
		System.out.println("**** genes ****");
		anal.printGenes();
	}

	public static Hashtable<String, HashSet<String>>  readGeneSets(String path, HashSet<String> bg) {
		BufferedReader reader;
		Hashtable<String, HashSet<String>>  geneSets = new Hashtable<String, HashSet<String>>();
		try {
			reader = new BufferedReader(new FileReader(path));
			String currLine = "";
			while ((currLine = reader.readLine()) != null) {
				String[] parsedLine = currLine.split("\t");
				if (parsedLine.length < 2) {
					continue;
				}
				String setName = parsedLine[1];
				if (setName.equals("grey")){
					continue;
				}
				String gene = parsedLine[0];
				if (! bg.contains(gene)) {
					continue;
				}

				if (!geneSets.containsKey(setName)) {
					HashSet<String> set = new HashSet<String>();
					set.add(gene);
					geneSets.put(setName, set);
				} else {
					geneSets.get(setName).add(gene);
				}

			}
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return geneSets;
	}
	
	public static Hashtable<String, HashSet<String>>  readGeneSets(String path, Collection<String> bg) {
		BufferedReader reader;
		Hashtable<String, HashSet<String>>  geneSets = new Hashtable<String, HashSet<String>>();
		try {
			reader = new BufferedReader(new FileReader(path));
			String currLine = "";
			while ((currLine = reader.readLine()) != null) {
				String[] parsedLine = currLine.split("\t");
				if (parsedLine.length < 2) {
					continue;
				}
				String setName = parsedLine[1];
				if (setName.equals("grey")){
					continue;
				}
				String gene = parsedLine[0];
				if (! bg.contains(gene)) {
					continue;
				}

				if (!geneSets.containsKey(setName)) {
					HashSet<String> set = new HashSet<String>();
					set.add(gene);
					geneSets.put(setName, set);
				} else {
					geneSets.get(setName).add(gene);
				}

			}
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return geneSets;
	}

}
